// dXView.h : interface of the CDXView class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_DXVIEW_H__E705677E_395D_4E5F_A397_36A7915F2EF2__INCLUDED_)
#define AFX_DXVIEW_H__E705677E_395D_4E5F_A397_36A7915F2EF2__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <atlmisc.h>
#include "DXSurfaceMgr.h"
#include <vector>
#include <algorithm>

enum DIRECTION
{
	UP,
	DOWN,
	DIAG,
};

enum SHAPE
{
	RECTNGLE,
	ELLIPSE,
};

const int move_amount = 1;

struct CMovingRect ;
struct ShapeDrawer
{
	virtual void Draw(const HDC& hdc, CMovingRect& mr)=0;
};

/*
	rect that moves up/down the client area
*/
struct CMovingRect : RECT
{
	HBRUSH clr;
	int dir;
	int ma;
	ShapeDrawer* _sd;

	CMovingRect(int l, int t, int r, int b, ShapeDrawer* sd)
		:_sd(sd)
	{ left =l; top = t; right = r; bottom = b; }
	
	~CMovingRect()
	{
		DeleteObject(clr);
		if(NULL!=_sd)
			delete _sd;
	}

	void Draw(const HDC& hdc)
	{
		_sd->Draw(hdc, *this);
	}

	void Move(const CRect& client)
	{
		if(this->top <=0)
			this->dir = DOWN;
		else if(this->bottom >= client.Height())
			this->dir = UP;

		switch(this->dir)
		{
			case UP:
			{
				this->top = this->top - this->ma;
				this->bottom = this->bottom - this->ma;
				break;
			}
			case DOWN:
			{
				this->top = this->top + this->ma;
				this->bottom = this->bottom + this->ma;
				break;
			}
		}
	}
};

struct ShapeDrawer_RC:ShapeDrawer
{
	void Draw(const HDC& hdc, CMovingRect& mr)
	{
		::FillRect(hdc, (RECT*)&mr, mr.clr);
	}
};

struct ShapeDrawer_E:ShapeDrawer
{
	void Draw(const HDC& hdc, CMovingRect& mr)
	{
		struct CAutoBM
		{
			HGDIOBJ _ob;
			const HDC& _hdc;
			CAutoBM(const HDC& hdc, HBRUSH hb)
				:_hdc(hdc)
			{
				 _ob=::SelectObject(hdc, hb);
			}
			~CAutoBM()
			{
				::SelectObject(_hdc, _ob);
			}

		} as(hdc, mr.clr);

		Ellipse(hdc, mr.left, mr.top, mr.right, mr.bottom); 
	}
};

struct MoveAndDrawOp
{
	const CRect& client_;
	const HDC& hdc_;

	MoveAndDrawOp(const HDC& hdc, const CRect& client)
	:client_(client), hdc_(hdc)
	{	}
	
	operator()(CMovingRect* mv)
	{
		mv->Move(client_);
		mv->Draw(hdc_);
	}
};

template <class T>
struct DeleteThing
{
	operator()(T* Thing)
	{
		delete Thing;
		Thing = NULL;
	}
};

typedef DeleteThing<CMovingRect> DeleteMovingRect;

class CDXView : public CWindowImpl<CDXView>
{
	std::vector<CMovingRect*> MovingRects;

	CDXSurfaceManager dxMgr_;
	class CPen
	{
		HPEN p_;
		HGDIOBJ op_;
		const HDC& hdc_;

	public:

		CPen(const HDC& hdc, UINT Style, UINT Width, int clr)
			:hdc_(hdc)
		{
			p_ = ::CreatePen(Style, Width, clr);
			op_ = ::SelectObject(hdc_, p_);
		}

		~CPen()
		{
			::SelectObject(hdc_, op_);	
			::DeleteObject(p_);
		}
	};

public:

	DECLARE_WND_CLASS(NULL)

	BOOL PreTranslateMessage(MSG* pMsg)
	{
		pMsg;
		return FALSE;
	}

	BEGIN_MSG_MAP(CDXView)
		MESSAGE_HANDLER(WM_PAINT, OnPaint)
		MESSAGE_HANDLER(WM_CREATE, OnCreate)
		MESSAGE_HANDLER(WM_ERASEBKGND, OnEraseBkgnd)
		MESSAGE_HANDLER(WM_TIMER, OnTimer)
		MESSAGE_HANDLER(WM_SIZE, OnSize)
		MESSAGE_HANDLER(WM_DESTROY, OnDestroy)
		MESSAGE_HANDLER(WM_SHOWWINDOW, OnShowWindow)
	END_MSG_MAP()

	LRESULT OnPaint(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
	{
		CRect client; GetClientRect(&client);

		// begin paint returns an hdc into which 
		// we can draw using normal gdi functions
		HDC hdc = dxMgr_.BeginPaint(GetSysColor(COLOR_WINDOW));
		ATLASSERT(hdc);

		CDCHandle cdc = hdc;
		// some normal gdi calls
		cdc.SetBkMode(TRANSPARENT);
		HBRUSH b = CreateSolidBrush(RGB(32,217,97));
		HBRUSH ob = cdc.SelectBrush(b);
		Ellipse(hdc, client.left, client.top, client.right, client.bottom); 
		cdc.SelectBrush(ob);
		
		filllines(hdc,client,RGB(27,03,65));

		DrawText(hdc,"Double buffering in DirectX", 28, &client, DT_CENTER|DT_VCENTER|DT_SINGLELINE);

		std::for_each(MovingRects.begin(), MovingRects.end(), MoveAndDrawOp(hdc, client));

		// End paint does the double buffering bit
		dxMgr_.EndPaint();
		return 0;
	}

	void filllines(HDC& hdc, const CRect& client, COLORREF clr = -1)
	{
		static int gap = 80;
		CPen p(hdc, PS_DASHDOTDOT, 1, clr);
		for (int i = gap ; i <= client.Height(); i = i+gap)
		{
			::MoveToEx(hdc, 0, i, NULL);
			::LineTo(hdc, client.Width(), i);
		}

		for (i = 80 ; i <= client.Width(); i = i+80)
		{
			::MoveToEx(hdc, i, 0, NULL);
			::LineTo(hdc, i, client.Height());
		}
	}

	LRESULT OnCreate(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		if(!dxMgr_.Initialise(m_hWnd)) return 0;
		return 0;
	}
	LRESULT OnEraseBkgnd(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		return 1;
	}
	LRESULT OnTimer(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		// first time init
		if(MovingRects.size()==0)
		{
			CRect client; GetClientRect(&client);
			for(int i = 0; i < 80;i++)
			{
				int pl = rand()%client.Width();
				int pt = rand()%client.Height();

				ShapeDrawer* sd = 0;
				int shape = rand()%2;
				switch (shape)
				{
				case RECTNGLE:
					{
						sd = new ShapeDrawer_RC;
					}
					break;
				case ELLIPSE:
					{
						sd = new ShapeDrawer_E;
					}
					break;
				}

				CMovingRect* mr = new CMovingRect(pl, pt, pl+(rand()%90), pt+(rand()%100), sd);
				mr->clr = ::CreateSolidBrush(RGB(rand()%255, rand()%255, rand()%255));
				mr->dir = rand()%2;
				mr->ma = (rand()%4)+move_amount;
				MovingRects.push_back(mr);
			}
		}
		Invalidate();
		UpdateWindow();
		return 0;
	}

	LRESULT OnSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		ClearMVs();
		return 0;
	}
	LRESULT OnDestroy(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		ClearMVs();
		return 0;
	}

	void ClearMVs()
	{
		std::for_each(MovingRects.begin(), MovingRects.end(), DeleteMovingRect()); 
		MovingRects.clear();
	}
	LRESULT OnShowWindow(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		SetTimer(1,33);		
		return 0;
	}
};


/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_DXVIEW_H__E705677E_395D_4E5F_A397_36A7915F2EF2__INCLUDED_)
